module Parser (Pos, Piece, piece, readPieces) where

import Char
import List

-- Errors handled: 
-- > Missing ':' in line
-- > No name before ':'
-- > error in coordinate list
-- > repeated name
-- > repeated coordinate in a piece

type Pos = (Int, Int)

newtype Piece = Piece (String, [Pos])

piece (Piece sp) = sp

isNumber ""  = False
isNumber ('+':str) = all isDigit str  
isNumber ('-':str) = all isDigit str  
isNumber str = all isDigit str  

trim =  reverse . dropWhile isSpace . reverse . dropWhile isSpace

-- File format: [PieceName: [(x,y)]+]+
parseLine (linenumber, filename) linetext 
  | null coordstring = fail ("Error in file '"++filename++"': Missing ':' in line "++show linenumber)
  | null name        = fail ("Error in file '"++filename++"': Unnamed piece in line "++show linenumber)
  | otherwise        = do { coords <- parseCoords coordstring'; return (Piece (name, coords)) }
  where
    (name',coordstring) = break (':'==) (trim linetext)
    name                = trim name'
    coordstring'        = filter (not . isSpace) $ tail coordstring

    parseCoord str 
      | length str < 5 || head str /= '(' || last str /= ')' || null ystr ||
        not (isNumber xstr) || not (isNumber ystr)  
             = fail ("Error in file '"++filename++"': Cannot parse coordinate list in line "++show linenumber)
      | otherwise = return (read xstr, read ystr)
      where
        (xstr0,ystr0) = break (','==) str
        xstr = tail xstr0
        ystr = init $ tail ystr0

    parseCoords "" = return []
    parseCoords str
      | null cs   = fail ("Error in file '"++filename++"': Cannot parse coordinate list in line "++show linenumber)
      | otherwise = do { cc <- parseCoord (c++")"); ccs <- parseCoords (tail cs); return (cc:ccs) }
      where 
        (c,cs) = break (')'==) str
        
        

readPieces :: String -> IO [Piece]
readPieces filename = 
  do { txt <- readFile filename
     ; lines <- parseLines [ (n,filename,l) | (n,l) <- zip [1..] (lines txt), not (null l) ]
     ; validate lines
     }
  where 
     parseLines [] = return []
     parseLines ((n,fn,l):rest) = 
       do { a <- parseLine (n,filename) l; as <- parseLines rest; return (a:as) }

     validate ps =
       do { norepeatedname names; norepeatedcoords pieces; return ps }
       where
         names  = map (fst . piece) ps

         norepeatedname [] = return True
         norepeatedname (n:ns) 
           | n `elem` ns = fail ("Error in file '"++filename++"': Repeated definition of piece named '"++n++"'")
           | otherwise   = norepeatedname ns

         pieces = map piece ps
         norepeatedcoords [] = return True
         norepeatedcoords ((p,cs):ps) 
           | length (nub cs) /= length cs = fail ("Error in file '"++filename++"': Repeated coordinate in piece '"++p++"'")
           | otherwise                    = norepeatedcoords ps
